/*
 *  Copyright 2009 RISC                            
 *  Copyright 2009 Karbin                          
 *  Distributed under the terms of the MIT License.
 */

module sample.chicago.crowd;

public import derelict.horde3d.horde3d;
public import derelict.horde3d.horde3dutils;

import tango.stdc.math;
import tango.stdc.stdlib;
import tango.util.container.more.Vector;

class Particle {

	float       px, pz;  // Current postition
	float       dx, dz;  // Destination position
	float       fx, fz;  // Force on particle
	float       ox, oz;  // Orientation vector
	NodeHandle  node;
	float       animTime;

	public this() {		
		ox = 0;
		oz = 0;
		node = 0;
		animTime = 0;
	}
	
}

class CrowdSim {

	Vector!(Particle, 100) particles;
	
	public this(char[] content_dir) {
		// Init random generator
		srand( 99777 );

		// Load character with walk animation
		ResHandle characterRes = Horde3D.addResource(ResourceTypes.SceneGraph, "man.scene.xml", 0);
		ResHandle characterWalkRes = Horde3D.addResource(ResourceTypes.Animation, "man.anim", 0);
		Horde3DUtils.loadResourcesFromDisk(content_dir.ptr);

		// Add characters
		for(uint i = 0; i < 100; ++i) {
			Particle p = new Particle();

		
			// Add character to scene and apply animation
			p.node = Horde3D.addNodes(RootNode, characterRes);
			Horde3D.setupModelAnimStage(p.node, 0, characterWalkRes, "", false);
		
			// Characters start in a circle formation
			p.px = sinf( (i / 100.0f) * 6.28f ) * 10.0f;
			p.pz = cosf( (i / 100.0f) * 6.28f ) * 10.0f;

			chooseDestination(p);

			Horde3D.setNodeTransform(p.node, p.px, 0.02f, p.pz, 0, 0, 0, 1, 1, 1);

			particles.add(p);
		}
	}
	
	public this() {
	}
	
	public void chooseDestination(Particle p) {
		// Choose random destination within a circle
		float ang = ((rand() % 360) / 360.0f) * 6.28f;
		float rad = (rand() % 20) * 1.0f;

		p.dx = sinf( ang ) * rad;
		p.dz = cosf( ang ) * rad;
	}

	public void update(float fps) {
		// Parameters for three repulsion zones
		float d1 = 0.25f, d2 = 2.0f, d3 = 4.5f;
		float f1 = 3.0f, f2 = 1.0f, f3 = 0.1f;
	
		for(int i = 0; i < particles.size() ; ++i) {
			Particle p = particles[i];
			// Reset force
			p.fx = 0; p.fz = 0;
		
			// Calculate distance to destination
			float dist = sqrtf( (p.dx - p.px)*(p.dx - p.px) + (p.dz - p.pz)*(p.dz - p.pz) );
		
			// If destination not reached walk to destination
			if( dist > 3.0f ) {
				// Calculate normalized attraction force to destination
				float afx = (p.dx - p.px) / dist;
				float afz = (p.dz - p.pz) / dist;	

				p.fx += afx * 0.035f; p.fz += afz * 0.035f;

				// Repulsion forces from other particles
				for( uint j = 0; j < particles.size(); ++j ) {
					if( j == i ) continue;
				
					Particle p2 = particles[j];
				
					float dist2 = sqrtf( (p.px - p2.px)*(p.px - p2.px) + (p.pz - p2.pz)*(p.pz - p2.pz) );
					float strength = 0;

					float rfx = (p.px - p2.px) / dist2;
					float rfz = (p.pz - p2.pz) / dist2;
				
					// Use three zones with different repulsion strengths
					if( dist2 <= d3 && dist2 > d2 ) {
						float m = (f3 - 0) / (d2 - d3);
						float t = 0 - m * d3;
						strength = m * dist2 + t;
					}
					
					else if( dist2 <= d2 && dist2 > d1 ) {
						float m = (f2 - f3) / (d1 - d2);
						float t = f3 - m * d2;
						strength = m * dist2 + t;
					}
					
					else if( dist2 <= d1 )	{
						float m = (f1 - f2) / (0 - d1);
						float t = f2 - m * d1;
						strength = m * dist2 + t;
					}

					p.fx += rfx * strength; p.fz += rfz * strength;
				}
			}
			
			else {
				// Choose new destination
				chooseDestination(p);
			}

			// Make movement frame rate independence
			p.fx *= (30 / fps);
			p.fz *= (30 / fps);
			float vel = sqrtf( p.fx * p.fx + p.fz * p.fz );
			
			// Set new position
			p.px += p.fx; p.pz += p.fz;
			
			// Caluclate orientation
			p.ox = (p.ox + p.fx) / 2;
			p.oz = (p.oz + p.fz) / 2;

			// Get rotation from orientation
			float ry = 0;
			if( p.oz != 0 ) ry = atan2( p.ox, p.oz );
			ry *= 180 / 3.1415f;	// Convert from radians to degrees
			
			// Update character scene node position
			Horde3D.setNodeTransform( p.node, p.px, 0.02f, p.pz, 0, ry, 0, 1, 1, 1 );
			
			// Update animation
			p.animTime += vel * 35.0f;
			Horde3D.setModelAnimParams( p.node, 0, p.animTime, 1.0f );
		}
	}
}
